/*
 * Copyright (C) 2006 Tilman Sauerbeck (tilman at code-monkey de)
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation, version 2.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301 USA
 */

#include <ruby.h>
#include <stdio.h>
#include <stdbool.h>
#include <errno.h>
#include <assert.h>

#include "vcedit.h"
#include "comments.h"

#define CHECK_CLOSED(o) \
	if (!o->state) \
		rb_raise (eClosed, "closed stream");

typedef struct {
	vcedit_state *state;
	VALUE comments;
} RbVorbisTagger;

static VALUE c_close (VALUE self);

static VALUE cComments, eVT, eClosed, eOpen, eInvalidData,
             eInvalidComment, eTempFile, eReopen;
static ID id_length;

static void
c_free (RbVorbisTagger *o)
{
	/* just in case the user forgot to call #close himself */
	if (o->state)
		vcedit_state_unref (o->state);

	ruby_xfree (o);
}

static void
c_mark (RbVorbisTagger *o)
{
	rb_gc_mark (o->comments);
}

static VALUE
c_alloc (VALUE klass)
{
	RbVorbisTagger *o;

	return Data_Make_Struct (klass, RbVorbisTagger, c_mark, c_free, o);
}

/*
 * call-seq:
 *  Ogg::Vorbis::Tagger.open(arg)                    -> object
 *  Ogg::Vorbis::Tagger.open(arg) { |object| block } -> object
 *
 * If a block isn't specified, Ogg::Vorbis::Tagger.open is a synonym
 * for Ogg::Vorbis::Tagger.new.
 * If a block is given, it will be invoked with the * Ogg::Vorbis::Tagger
 * object as a parameter.
 */
static VALUE
c_open (VALUE klass, VALUE arg)
{
	VALUE obj = rb_class_new_instance (1, &arg, klass);

	if (rb_block_given_p ())
		return rb_ensure (rb_yield, obj, c_close, obj);
	else
		return obj;
}

/*
 * call-seq:
 *  Ogg::Vorbis::Tagger.new(filename) -> object
 *
 * Returns a new Ogg::Vorbis::Tagger object for the specified file.
 *
 * FIXME: add optional mode argument (read-only or read-write)
 */
static VALUE
c_init (VALUE self, VALUE filename)
{
	RbVorbisTagger *o;
	vorbis_comment *vc;
	int i;

	Data_Get_Struct (self, RbVorbisTagger, o);

	StringValue (filename);

	o->state = vcedit_state_new (StringValuePtr (filename));
	if (!o->state)
		rb_raise (rb_eNoMemError, "Out of Memory");

	switch (vcedit_open (o->state)) {
		case VCEDIT_ERR_OPEN:
			rb_raise (eOpen, "Cannot open file");
		case VCEDIT_ERR_INVAL:
			rb_raise (eInvalidData, "Invalid data");
		default:
			break;
	}

	vc = vcedit_comments (o->state);

	/* vcedit_open() succeeded, so vcedit_comments() cannot
	 * return NULL.
	 */
	assert (vc);

	/* check whether all comments are well-formed */
	for (i = 0; i < vc->comments; i++) {
		char *ptr, *content = vc->user_comments[i];

		ptr = strchr (content, '=');
		if (!ptr || ptr == content)
			rb_raise (eInvalidComment, "invalid comment - %s", content);
	}

	o->comments = rb_class_new_instance (0, NULL, cComments);

	comments_init (o->comments, o->state);

	return self;
}

/*
 * call-seq:
 *  object.close -> nil
 *
 * Closes *object*. Further method calls on *object* will raise an
 * Ogg::Vorbis::Tagger::ClosedStreamError exception.
 * Returns +nil+.
 */
static VALUE
c_close (VALUE self)
{
	RbVorbisTagger *o;

	Data_Get_Struct (self, RbVorbisTagger, o);

	CHECK_CLOSED (o);

	vcedit_state_unref (o->state);
	o->state = NULL;

	return Qnil;
}

/*
 * call-seq:
 *  object.write -> integer
 *
 * Writes the comments from *object* back to the IO object and
 * returns the numbers of comments written.
 */
static VALUE
c_write (VALUE self)
{
	RbVorbisTagger *o;

	Data_Get_Struct (self, RbVorbisTagger, o);

	CHECK_CLOSED (o);

	comments_sync (o->comments, o->state);

	switch (vcedit_write (o->state)) {
		case VCEDIT_ERR_INVAL:
			rb_raise (eInvalidData, "Invalid data");
		case VCEDIT_ERR_TMPFILE:
			rb_raise (eTempFile, "Cannot create temporary file");
		case VCEDIT_ERR_REOPEN:
			rb_raise (eReopen, "Cannot reopen file");
		default:
			break;
	}

	return rb_funcall (o->comments, id_length, 0);
}

/*
 * call-seq:
 *  object.comments -> comments
 *
 * Returns the comments collection of *object*, which is an instance of
 * Ogg::Vorbis::Comments.
 */
static VALUE
c_comments (VALUE self)
{
	RbVorbisTagger *o;

	Data_Get_Struct (self, RbVorbisTagger, o);

	CHECK_CLOSED (o);

	return o->comments;
}

EXT_API
void
Init_vorbistagger_ext (void)
{
	VALUE mOgg, mVorbis, eOgg, eVorbis, cVT;

	mOgg = rb_define_module ("Ogg");
	mVorbis = rb_define_module_under (mOgg, "Vorbis");

	eOgg = rb_define_class_under (mOgg, "OggError", rb_eStandardError);
	eVorbis = rb_define_class_under (mVorbis, "VorbisError", eOgg);

	cVT = rb_define_class_under (mVorbis, "Tagger", rb_cObject);

	rb_define_alloc_func (cVT, c_alloc);

	rb_define_singleton_method (cVT, "open", c_open, 1);
	rb_define_method (cVT, "initialize", c_init, 1);
	rb_define_method (cVT, "close", c_close, 0);
	rb_define_method (cVT, "write", c_write, 0);
	rb_define_method (cVT, "comments", c_comments, 0);

	eVT = rb_define_class_under (cVT, "TaggerError", eVorbis);
	eClosed = rb_define_class_under (cVT, "ClosedStreamError", eVT);
	eOpen = rb_define_class_under (cVT, "OpenError", eVT);
	eInvalidData = rb_define_class_under (cVT, "InvalidDataError", eVT);
	eInvalidComment = rb_define_class_under (cVT, "InvalidCommentError",
	                                         eInvalidData);
	eTempFile = rb_define_class_under (cVT, "TempFileError", eVT);
	eReopen = rb_define_class_under (cVT, "ReopenError", eVT);

	id_length = rb_intern ("length");

	cComments = Init_Comments (mVorbis);
}
